package org.xesstandard.xml;

import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.UUID;

import javax.xml.bind.DatatypeConverter;

import org.xesstandard.factory.XesFactory;
import org.xesstandard.factory.impl.XesFactoryImpl;
import org.xesstandard.model.XesAttribute;
import org.xesstandard.model.XesClassifier;
import org.xesstandard.model.XesComponent;
import org.xesstandard.model.XesEvent;
import org.xesstandard.model.XesExtension;
import org.xesstandard.model.XesExtensionManager;
import org.xesstandard.model.XesGlobal;
import org.xesstandard.model.XesLog;
import org.xesstandard.model.XesTrace;
import org.xesstandard.model.XesValues;
import org.xesstandard.model.attributes.XesBooleanAttribute;
import org.xesstandard.model.attributes.XesDateTimeAttribute;
import org.xesstandard.model.attributes.XesIdAttribute;
import org.xesstandard.model.attributes.XesIntegerNumberAttribute;
import org.xesstandard.model.attributes.XesListAttribute;
import org.xesstandard.model.attributes.XesRealNumberAttribute;
import org.xesstandard.model.attributes.XesStringAttribute;
import org.xesstandard.xml.issues.XesXmlIssueAttributeGlobal;
import org.xesstandard.xml.issues.XesXmlIssueAttributePrefix;
import org.xesstandard.xml.issues.XesXmlIssueAttributeRequired;
import org.xesstandard.xml.issues.XesXmlIssueAttributeType;
import org.xesstandard.xml.issues.XesXmlIssueAttributeUnique;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueBoolean;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueDate;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueDecimal;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueDouble;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueEventTrace;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueLong;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueNCName;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueTZD;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValueUUID;
import org.xesstandard.xml.issues.XesXmlIssueAttributeValues;
import org.xesstandard.xml.issues.XesXmlIssueClassifierGlobal;
import org.xesstandard.xml.issues.XesXmlIssueClassifierNoKeys;
import org.xesstandard.xml.issues.XesXmlIssueExtensionKey;
import org.xesstandard.xml.issues.XesXmlIssueLogRoot;
import org.xesstandard.xml.issues.XesXmlIssueNestedAttribute;
import org.xesstandard.xml.issues.XesXmlIssueNoNestedAttribute;
import org.xesstandard.xml.issues.XesXmlIssueOutOfPlace;
import org.xesstandard.xml.issues.XesXmlIssueParent;
import org.xesstandard.xml.issues.XesXmlIssueStrictlyConformingAttribute;
import org.xesstandard.xml.issues.XesXmlIssueStrictlyConformingElement;
import org.xesstandard.xml.issues.XesextXmlIssueParser;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class XesXmlHandler extends DefaultHandler {

	/*
	 * The lowest issue level that is considered to be an error. All issues
	 * detected below this level will be considered to be mere warnings, all
	 * issues from this level up will be considered to be errors. If any errors
	 * are detected, <null> will be returned for the log by the parser.
	 */
	private XesXmlIssueLevel lowestErrorLevel;

	/*
	 * XES Factory to use.
	 */
	private XesFactory factory;

	/*
	 * The log.
	 */
	private XesLog log;

	/*
	 * The current trace.
	 */
	private XesTrace trace;

	/*
	 * The current list of global attributes.
	 */
	private List<XesAttribute> globals;

	/*
	 * A stack containing all current list attributes.
	 */
	private Stack<XesListAttribute> lists;

	/*
	 * State counter for every list on the previous stack. Counter is
	 * incremented whenever we encounter a values start or end tag.
	 */
	private Stack<Integer> inValues;

	/*
	 * Counter for number of globals we have seen so far.
	 */
	private int globalCtr;

	/*
	 * Stack of current components.
	 */
	private Stack<XesComponent> components;

	/*
	 * Set of attribute keys that are used in classifiers. For certain attribute
	 * types (string, id, date), it is worthwhile to not create duplicate values
	 * for attributes with these keys.
	 */
	private Set<String> classifiedKeys;

	/*
	 * A map containing all keys.
	 */
	private Map<String, String> allKeys;

	/*
	 * A map containing all classified string values. This prevents these string
	 * values to be kept more than once in memory.
	 */
	private Map<String, String> classifiedStringValues;

	/*
	 * A map containing all classified date values. This prevents these date
	 * values to be kept more than once in memory.
	 */
	private Map<Date, Date> classifiedDateValues;

	/*
	 * A map containing all classified id values. This prevents these id values
	 * to be kept more than once in memory.
	 */
	private Map<UUID, UUID> classifiedUUIDValues;

	/*
	 * The list of standard extensions.
	 */
	private XesExtensionManager extensionManager;

	/*
	 * Dummy extension to be used for attributes that have no prefix in their
	 * key.
	 */
	private XesExtension noExtension;

	/*
	 * Whether or not the log declares the nested-attributes feature.
	 */
	private boolean hasNestedAttributes;

	/*
	 * The list of issues raised while parsing.
	 */
	private List<XesXmlIssue> issues;

	public XesXmlHandler(XesXmlIssueLevel level) {
		/*
		 * Set the lowest warning level.
		 */
		lowestErrorLevel = level;
		/*
		 * Initialize the rest.
		 */
		factory = new XesFactoryImpl();
		log = null;
		trace = null;
		globals = null;
		globalCtr = 0;
		lists = new Stack<XesListAttribute>();
		inValues = new Stack<Integer>();
		components = new Stack<XesComponent>();
		classifiedKeys = new HashSet<String>();
		allKeys = new HashMap<String, String>();
		classifiedStringValues = new HashMap<String, String>();
		classifiedDateValues = new HashMap<Date, Date>();
		classifiedUUIDValues = new HashMap<UUID, UUID>();
		extensionManager = XesExtensionManager.instance();
		noExtension = factory.createExtension();
		hasNestedAttributes = false;
		issues = new ArrayList<XesXmlIssue>();
	}

	/*
	 * Locator to be able to store line and column numbers while parsing.
	 */
	private Locator locator;

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#setDocumentLocator(org.xml.sax.
	 * Locator)
	 */
	public void setDocumentLocator(Locator locator) {
		this.locator = locator;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String,
	 * java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {

		// Preprocess tag name.
		String tagName = localName.trim();
		if (tagName.length() == 0) {
			tagName = qName;
		}

		// Check tag name.
		if (tagName.equals(XesXmlTags.STRING)) {
			// Start a string attribute.
			XesStringAttribute attribute = factory.createStringAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
		} else if (tagName.equals(XesXmlTags.DATE)) {
			// Start a date attribute.
			XesDateTimeAttribute attribute = factory.createDateTimeAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
		} else if (tagName.equals(XesXmlTags.INT)) {
			// Start an integer number attribute.
			XesIntegerNumberAttribute attribute = factory.createIntegerNumberAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
		} else if (tagName.equals(XesXmlTags.FLOAT)) {
			// Start a real number attribute.
			XesRealNumberAttribute attribute = factory.createRealNumberAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
		} else if (tagName.equals(XesXmlTags.BOOLEAN)) {
			// Start a boolean attribute.
			XesBooleanAttribute attribute = factory.createBooleanAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
		} else if (tagName.equals(XesXmlTags.ID)) {
			// Start an id attribute.
			XesIdAttribute attribute = factory.createIdAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
		} else if (tagName.equals(XesXmlTags.LIST)) {
			// Start a list attribute.
			XesListAttribute attribute = factory.createListAttribute();
			startAttributeElement(attribute, attributes);
			components.push(attribute);
			// Additional administration for lists.
			lists.push(attribute);
			inValues.push(0);
		} else if (tagName.equals(XesXmlTags.VALUES)) {
			// Start of a values list. 
			XesValues values = factory.createValues();
			startValuesElement(values, attributes);
			// Increase the inValues counter for this list.
			inValues.push(inValues.pop() + 1);
		} else if (tagName.equals(XesXmlTags.EVENT)) {
			// Start of an event.
			XesEvent event = factory.createEvent();
			startEventElement(event, attributes);
			components.push(event);
		} else if (tagName.equals(XesXmlTags.TRACE)) {
			// Start of a trace.
			XesTrace trace = factory.createTrace();
			startTraceElement(trace, attributes);
			components.push(trace);
		} else if (tagName.equals(XesXmlTags.LOG)) {
			// Start of a log.
			XesLog log = factory.createLog();
			startLogElement(log, attributes);
			components.push(log);
		} else if (tagName.equals(XesXmlTags.EXTENSION)) {
			// Start of an extension.
			XesExtension extension = factory.createExtension();
			startExtensionElement(extension, attributes);
		} else if (tagName.equals(XesXmlTags.CLASSIFIER)) {
			// Start of a classifier.
			XesClassifier classifier = factory.createClassifier();
			startClassifierElement(classifier, attributes);
		} else if (tagName.equals(XesXmlTags.GLOBAL)) {
			// Start of a global.
			XesGlobal global = factory.createGlobal();
			startGlobalElement(global, attributes);
		} else {
			// Unknown start tag. Raise issue.
			XesXmlIssue issue = new XesXmlIssueStrictlyConformingElement(tagName);
			issue.setLocation(locator);
			issues.add(issue);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String,
	 * java.lang.String, java.lang.String)
	 */
	public void endElement(String uri, String localName, String qName) throws SAXException {

		// Preprocess tag name.
		String tagName = localName.trim();
		if (tagName.length() == 0) {
			tagName = qName;
		}

		// Check tag name.
		if (tagName.equals(XesXmlTags.STRING)) {
			// End string attribute.
			components.pop();
		} else if (tagName.equals(XesXmlTags.DATE)) {
			// End date attribute.
			components.pop();
		} else if (tagName.equals(XesXmlTags.INT)) {
			// End integer number attribute.
			components.pop();
		} else if (tagName.equals(XesXmlTags.FLOAT)) {
			// End real number attribute.
			components.pop();
		} else if (tagName.equals(XesXmlTags.BOOLEAN)) {
			// End boolean attribute.
			components.pop();
		} else if (tagName.equals(XesXmlTags.ID)) {
			// End id attribute.
			components.pop();
		} else if (tagName.equals(XesXmlTags.LIST)) {
			// End list attribute.
			components.pop();
			// Additional administration for list.
			if (inValues.peek() <= 0) {
				// No values tag seen for this list. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValues(lists.peek());
				issue.setLocation(locator);
				issues.add(issue);
			}
			lists.pop();
			inValues.pop();
		} else if (tagName.equals(XesXmlTags.VALUES)) {
			// End of values for list.
			inValues.push(inValues.pop() + 1);
		} else if (tagName.equals(XesXmlTags.EVENT)) {
			// End of event.
			XesEvent event = (XesEvent) components.pop();
			checkGlobals(event);
		} else if (tagName.equals(XesXmlTags.TRACE)) {
			// End of trace.
			components.pop(); // must be this.trace
			endTraceElement(trace);
		} else if (tagName.equals(XesXmlTags.LOG)) {
			// End of log.
			if (log != null && log.getFeatures().contains(XesXmlTags.NESTEDATTRIBUTES) && !hasNestedAttributes) {
				// Log declares nested attributes, but does not have any nested attributes. Raise issue.
				XesXmlIssue issue = new XesXmlIssueNoNestedAttribute(log);
				issue.setLocation(locator);
				issues.add(issue);
			}
			components.pop(); // must be this.log
		} else if (tagName.equals(XesXmlTags.EXTENSION)) {
			// End of extension.
		} else if (tagName.equals(XesXmlTags.CLASSIFIER)) {
			// End of classifier.
		} else if (tagName.equals(XesXmlTags.GLOBAL)) {
			// End of global.
			globals = null;
		}

	}

	/*
	 * Start a log.
	 */
	private void startLogElement(XesLog log, Attributes attributes) {

		if (!components.isEmpty()) {
			// Log is not root. Raise issue.
			XesXmlIssue issue = new XesXmlIssueLogRoot(log);
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Handle required decimal "xes.version" attribute.
		String value = attributes.getValue(XesXmlTags.XESVERSION);
		if (value == null) {
			// Attribute missing. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(log, XesXmlTags.XESVERSION);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			Double decimal = getDecimal(value);
			if (decimal == null) {
				// Attribute is not decimal. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueDecimal(log, XesXmlTags.XESVERSION, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				log.setVersion(value);
			}
		}

		// Handle required string "xes.features" attribute.
		value = attributes.getValue(XesXmlTags.XESFEATURES);
		if (value == null) {
			// Attribute missing. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(log, XesXmlTags.XESFEATURES);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			// Attribute must be white-space separated list of features.
			String[] features = value.split(" ");
			for (int i = 0; i < features.length; i++) {
				if (!features[i].isEmpty()) {
					// Feature found. Add it.
					log.addFeature(features[i]);
				}
			}
		}

		// Handle additional attributes. 
		for (int i = 0; i < attributes.getLength(); i++) {
			value = attributes.getQName(i);
			if (value.equals(XesXmlTags.XESVERSION) || value.equals(XesXmlTags.XESFEATURES)) {
			} else {
				// Unknown attribute. Raise issue.
				XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(log, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Set first log we encounter.
		if (this.log == null) {
			this.log = log;
		}
	}

	/*
	 * Start a trace.
	 */
	private void startTraceElement(XesTrace trace, Attributes attributes) {

		if (components.isEmpty() || components.peek() != log) {
			// Parent is not a log. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(trace, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			// Unknown attribute. Raise issue.
			String value = attributes.getQName(i);
			XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(trace, value);
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (!log.getEvents().isEmpty()) {
			// Trace must precede any event. Raise issue.
			XesXmlIssue issue = new XesXmlIssueOutOfPlace(trace);
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (this.trace == null) {
			// Set current trace, and add it to log.
			this.trace = trace;
			log.addTrace(this.trace);
		}
	}

	/*
	 * End a trace.
	 */
	private void endTraceElement(XesTrace trace) {
		// Check trace for global attributes.
		checkGlobals(log.getGlobalTraceAttributes(), trace);
		// Unset current trace.
		this.trace = null;
	}

	/*
	 * Start an event.
	 */
	private void startEventElement(XesEvent event, Attributes attributes) {

		if (components.isEmpty() || (components.peek() != log && components.peek() != trace)) {
			// Parent is not a log and not a trace. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(event, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			// Unknown attribute. Raise issue.
			String value = attributes.getQName(i);
			XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(event, value);
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (components.peek() == trace) {
			// Parent is current trace. Add event to current trace.
			trace.addEvent(event);
		} else {
			// Parent must be log. Add event to log.
			log.addEvent(event);
		}
	}

	/*
	 * End an event.
	 */
	private void checkGlobals(XesEvent event) {
		// Check event for global attributes.
		checkGlobals(log.getGlobalEventAttributes(), event);
	}

	/*
	 * Start an extension.
	 */
	private void startExtensionElement(XesExtension extension, Attributes attributes) {

		if (components.isEmpty() || components.peek() != log) {
			// Parent is not a log. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(extension, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (globalCtr > 0 || !log.getEventClassifiers().isEmpty() || !log.getTraceClassifiers().isEmpty()
				|| !log.getAttributes().isEmpty() || !log.getTraces().isEmpty() || !log.getEvents().isEmpty()) {
			// Extension must precede any global, any classifier, any attribute, any trace, and any event. Raise issue.
			XesXmlIssue issue = new XesXmlIssueOutOfPlace(extension);
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Handle extension name.
		String value = attributes.getValue(XesXmlTags.NAME);
		if (value == null) {
			// Extension lacks name attribute. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(extension, XesXmlTags.NAME);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			String nameValue = getNCName(value);
			if (nameValue == null) {
				// Name is not an NCName. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueNCName(extension, XesXmlTags.NAME, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				// Set extension name.
				extension.setName(nameValue);
			}
		}

		// Handle extension prefix.
		value = attributes.getValue(XesXmlTags.PREFIX);
		if (value == null) {
			// Extension lack prefix attribute. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(extension, XesXmlTags.PREFIX);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			String nameValue = getNCName(value);
			if (nameValue == null) {
				// Prefix is not an NCName. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueNCName(extension, XesXmlTags.PREFIX, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				// Set extension prefix.
				extension.setPrefix(value);
			}
		}

		// Handle extension URI.
		value = attributes.getValue(XesXmlTags.URI);
		if (value == null) {
			// Extension lacks uri attribute. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(extension, XesXmlTags.URI);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			URI uri = getURI(value);
			if (uri == null) {
				// Uri is not a valid URI. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueNCName(extension, XesXmlTags.URI, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				// Set extension URI.
				extension.setURI(uri);
			}
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			value = attributes.getQName(i);
			if (value.equals(XesXmlTags.NAME) || value.equals(XesXmlTags.PREFIX) || value.equals(XesXmlTags.URI)) {
			} else {
				// Unknown attribute. Raise issue.
				XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(extension, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Check whether extension matches existing extension.
		XesExtension existingExtension = extensionManager.getExtension(extension.getName(), extension.getPrefix(),
				extension.getURI());
		if (existingExtension != null) {
			// Yes it does, use existing extension.
			extension = existingExtension;
		} else {
			// No, it does not. Add it.
			extensionManager.addExtension(extension);			
		}

		if (extension.getEventAttributes().isEmpty() && extension.getTraceAttributes().isEmpty()
				&& extension.getLogAttributes().isEmpty() && extension.getMetaAttributes().isEmpty()) {
			// Load attribute information from XESEXT file.
			XesextXmlParser parser = new XesextXmlParser();
			try {
				parser.parse(extension, issues);
			} catch (Exception e) {
				XesXmlIssue issue = new XesextXmlIssueParser(extension, e.getMessage());
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Add extension to the log.
		log.addExtension(extension);
	}

	/*
	 * Start a global
	 */
	private void startGlobalElement(XesGlobal global, Attributes attributes) {

		if (components.isEmpty() || components.peek() != log) {
			// Paren tis not a log. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(global, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (!log.getEventClassifiers().isEmpty() || !log.getTraceClassifiers().isEmpty()
				|| !log.getAttributes().isEmpty() || !log.getTraces().isEmpty() || !log.getEvents().isEmpty()) {
			// Global must precede any classifier, any attribute, any trace, and any event. Raise issue.
			XesXmlIssue issue = new XesXmlIssueOutOfPlace(global);
			issue.setLocation(locator);
			issues.add(issue);
		}

		// By default, a global is an event global.
		List<XesAttribute> globals = log.getGlobalEventAttributes();

		String value = attributes.getValue(XesXmlTags.SCOPE);
		if (value == null) {
		} else {
			if (value.equals(XesXmlTags.EVENT)) {
			} else if (value.equals(XesXmlTags.TRACE)) {
				// Global is a trace global.
				globals = log.getGlobalTraceAttributes();
			} else {
				// Unknown value for scope. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueEventTrace(global, XesXmlTags.SCOPE, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			value = attributes.getQName(i);
			if (value.equals(XesXmlTags.SCOPE)) {
			} else {
				// Unknown attribute. Raise issue.
				XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(global, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Set globals list to correct globals list (either event or trace).
		this.globals = globals;

		// Advance global counter to indicate that we have seen some global.
		globalCtr++;
	}

	/*
	 * Start a classifier.
	 */
	private void startClassifierElement(XesClassifier classifier, Attributes attributes) {

		if (components.isEmpty() || components.peek() != log) {
			// Parent is not a log. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(classifier, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (!log.getAttributes().isEmpty() || !log.getTraces().isEmpty() || !log.getEvents().isEmpty()) {
			// Classifier must precede any attribute, any trace, and any event. Raise issue.
			XesXmlIssue issue = new XesXmlIssueOutOfPlace(classifier);
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Handle classifier name.
		String value = attributes.getValue(XesXmlTags.NAME);
		if (value == null) {
			// Classifier has no name attribute. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(classifier, XesXmlTags.NAME);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			String nameValue = getNCName(value);
			if (nameValue == null) {
				// Name is not an NCName. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueNCName(classifier, XesXmlTags.NAME, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				// Set classifier name.
				classifier.setName(value);
			}
		}

		// Handle classifier scope. By default, the classifier is an event classifier, and the keys
		// should all be event globals.
		List<XesClassifier> classifiers = log.getEventClassifiers();
		List<XesAttribute> globals = log.getGlobalEventAttributes();
		value = attributes.getValue(XesXmlTags.SCOPE);
		if (value != null) {
			if (value.equals(XesXmlTags.EVENT)) {
			} else if (value.equals(XesXmlTags.TRACE)) {
				// Trace classifier, keys should all be event globals.
				classifiers = log.getTraceClassifiers();
				globals = log.getGlobalEventAttributes();
			} else {
				// Unknown scope. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeValueEventTrace(classifier, XesXmlTags.SCOPE, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Handle classifier keys.
		value = attributes.getValue(XesXmlTags.KEYS);
		if (value == null) {
			// No keys. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(classifier, XesXmlTags.KEYS);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			String[] keys = value.split(" ");
			if (keys.length == 0 || keys[0].isEmpty()) {
				// Empty keys. Raise issue.
				XesXmlIssue issue = new XesXmlIssueClassifierNoKeys(classifier, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				// Try to fix the keys using the globals.
				List<String> fixedKeys = fixKeys(globals, Arrays.asList(keys));
				// Collect all keys used by the classifier.
				Set<String> globalKeys = new HashSet<String>();
				for (XesAttribute global : globals) {
					globalKeys.add(global.getKey());
				}
				if (!globalKeys.containsAll(fixedKeys)) {
					// Not all keys are global. Raise issue.
					Set<String> nonGlobalKeys = new HashSet<String>(fixedKeys);
					nonGlobalKeys.removeAll(globalKeys);
					XesXmlIssue issue = new XesXmlIssueClassifierGlobal(classifier, nonGlobalKeys);
					issue.setLocation(locator);
					issues.add(issue);
				}
				// Add the keys to the classifier. Also store the fact that this key is used in a classifier.
				for (String key : fixedKeys) {
					classifier.addKey(key);
					classifiedKeys.add(key);
				}
			}
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			value = attributes.getQName(i);
			if (value.equals(XesXmlTags.NAME) || value.equals(XesXmlTags.SCOPE) || value.equals(XesXmlTags.KEYS)) {
			} else {
				// Unknown attribute. Raise issue.
				XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(classifier, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Add classifier to correct list (either event or trace) of classifiers.
		classifiers.add(classifier);
	}

	/*
	 * Start an attribute.
	 */
	private void startAttributeElement(XesAttribute attribute, Attributes attributes) {

		if (globals == null && (inValues.isEmpty() || !(inValues.peek() % 2 == 1)) && components.isEmpty()) {
			// Parent must be a global, a list values, or a component. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(attribute, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}

		if ((inValues.isEmpty() || !(inValues.peek() % 2 == 1)) && !components.isEmpty()
				&& components.peek() instanceof XesAttribute) {
			if (!log.getFeatures().contains(XesXmlTags.NESTEDATTRIBUTES)) {
				// A nested attribute, but this feature is not declared by the log., Raise issue.
				XesXmlIssue issue = new XesXmlIssueNestedAttribute((XesAttribute) components.peek());
				issue.setLocation(locator);
				issues.add(issue);
			}
			hasNestedAttributes = true;
		}

		if (globals == null && (inValues.isEmpty() || !(inValues.peek() % 2 == 1)) && !components.isEmpty()
				&& components.peek() instanceof XesLog) {
			// A log attribute.
			if (!log.getTraces().isEmpty() || !log.getEvents().isEmpty()) {
				// A log attribute must precede any trace and any event. Raise issue.
				XesXmlIssue issue = new XesXmlIssueOutOfPlace(attribute, log.getTraces(), log.getEvents());
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		if (globals == null && (inValues.isEmpty() || !(inValues.peek() % 2 == 1)) && !components.isEmpty()
				&& components.peek() instanceof XesTrace) {
			// A trace attribute.
			if (!trace.getEvents().isEmpty()) {
				// A trace attribute should preced any event. Raise issue.
				XesXmlIssue issue = new XesXmlIssueOutOfPlace(attribute, trace.getEvents());
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		if ((inValues.isEmpty() || !(inValues.peek() % 2 == 1)) && !components.isEmpty()
				&& components.peek() instanceof XesListAttribute) {
			// A list attribute.
			XesListAttribute list = (XesListAttribute) components.peek();
			if (!inValues.isEmpty() && inValues.peek() > 0) {
				// A list attribute should precede any values. Raise issue.
				XesXmlIssue issue = new XesXmlIssueOutOfPlace(list);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		// Handle attribute key.
		String value = attributes.getValue(XesXmlTags.KEY);
		if (value == null) {
			// No key. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeRequired(attribute, XesXmlTags.KEY);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			// Check whether known key. Prevents having to keep multiple copies of same key.
			String key = allKeys.get(value);
			if (key == null) {
				// No, add to known keys.
				allKeys.put(value, value);
				key = value;
			}
			// Set known attribute key. 
			attribute.setKey(key);
		}

		// Handle extension for attribute.
		if (value != null) {
			// Get the extension based on the key prefix. Returns noExtension if no prefix, and
			// <null? if non-empty prefix which does not match any known extension.
			XesExtension extension = getExtension(value);
			if (extension == null) {
				// Unknown extension. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributePrefix(attribute, value);
				issue.setLocation(locator);
				issues.add(issue);
			} else if (extension != noExtension) {
				// Set attribute extension.
				attribute.setExtension(extension);
				// Check whether extension knows this key.
				boolean ok = false;
				String level = "<unknown>";
				if (globals == log.getGlobalTraceAttributes()) {
					ok = extension.hasTraceAttribute(value);
					level = "trace";
				} else if (globals == log.getGlobalEventAttributes()) {
					ok = extension.hasEventAttribute(value);
					level = "event";
				} else if (!inValues.isEmpty()) {
					ok = extension.hasMetaAttribute(value);
					level = "meta";
				} else if (components.isEmpty()) {
					// Should not be.
				} else if (components.peek() instanceof XesLog) {
					ok = extension.hasLogAttribute(value);
					level = "log";
				} else if (components.peek() instanceof XesTrace) {
					ok = extension.hasTraceAttribute(value);
					level = "trace";
				} else if (components.peek() instanceof XesEvent) {
					ok = extension.hasEventAttribute(value);
					level = "event";
				} else if (components.peek() instanceof XesAttribute) {
					ok = extension.hasMetaAttribute(value);
					level = "meta";
				}
				if (!ok) {
					// Extension does not know this key. Raise issue.
					XesXmlIssue issue = new XesXmlIssueExtensionKey(extension, value, level);
					issue.setLocation(locator);
					issues.add(issue);
				}
			}
		}

		// Handle simple attribute value
		if (attribute.isSimple()) {
			value = attributes.getValue(XesXmlTags.VALUE);
			if (value == null) {
				// No value. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeRequired(attribute, XesXmlTags.VALUE);
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				switch (attribute.getType()) {
					case STRING :
						if (classifiedKeys.contains(attribute.getKey())) {
							// Claasified key. Use map with known values.
							String mappedValue = classifiedStringValues.get(value);
							if (mappedValue == null) {
								// Unknown value. Add to map.
								classifiedStringValues.put(value, value);
							} else {
								// Known value. Use it instead.
								value = mappedValue;
							}
						}
						// Set string value for attribute.
						((XesStringAttribute) attribute).setValue(value);
						break;
					case BOOLEAN :
						Boolean b = getBoolean(value);
						if (b == null) {
							// Not a valid boolean value. Raise issue.
							XesXmlIssue issue = new XesXmlIssueAttributeValueBoolean(attribute, attribute.getKey(),
									value);
							issue.setLocation(locator);
							issues.add(issue);
						}
						// Set boolean value for attribute.
						((XesBooleanAttribute) attribute).setValue(b);
						break;
					case DATETIME :
						String tzd = getTZD(value);
						if (tzd == null) {
							// Not a valid time zone. Raise issue.
							XesXmlIssue issue = new XesXmlIssueAttributeValueTZD(attribute, attribute.getKey(), value);
							issue.setLocation(locator);
							issues.add(issue);
						}
						Date date = getDateTime(value);
						if (date == null) {
							// No valid UTC date. Raise issue.
							XesXmlIssue issue = new XesXmlIssueAttributeValueDate(attribute, attribute.getKey(), value);
							issue.setLocation(locator);
							issues.add(issue);
						}
						if (date != null && classifiedKeys.contains(attribute.getKey())) {
							// Claasified key. Use map with known values.
							Date mappedDate = classifiedDateValues.get(date);
							if (mappedDate == null) {
								// Unknown value. Add it.
								classifiedDateValues.put(date, date);
							} else {
								// Known value. Use it instead.
								date = mappedDate;
							}
						}
						// Set UTC date value for attribute.
						((XesDateTimeAttribute) attribute).setValue(date);
						break;
					case ID :
						UUID id = getId(value);
						if (id == null) {
							// Not a valid UUID. Raise issue.
							XesXmlIssue issue = new XesXmlIssueAttributeValueUUID(attribute, attribute.getKey(), value);
							issue.setLocation(locator);
							issues.add(issue);
						}
						if (classifiedKeys.contains(attribute.getKey())) {
							// Classified key. Use map with known values.
							UUID mappedId = classifiedUUIDValues.get(id);
							if (mappedId == null) {
								// Unknown value. Add it.
								classifiedUUIDValues.put(id, id);
							} else {
								// Known value. Use it instead.
								id = mappedId;
							}
						}
						// Set UUID value for attribute.
						((XesIdAttribute) attribute).setValue(id);
						break;
					case INTEGERNUMBER :
						Long l = getLong(value);
						if (l == null) {
							// Not a valid integer number. Raise issue.
							XesXmlIssue issue = new XesXmlIssueAttributeValueLong(attribute, attribute.getKey(), value);
							issue.setLocation(locator);
							issues.add(issue);
						}
						// Set integer number value for attribute.
						((XesIntegerNumberAttribute) attribute).setValue(l);
						break;
					case REALNUMBER :
						Double d = getDouble(value);
						if (d == null) {
							// Not a valid rela number. Raise issue.
							XesXmlIssue issue = new XesXmlIssueAttributeValueDouble(attribute, attribute.getKey(),
									value);
							issue.setLocation(locator);
							issues.add(issue);
						}
						// Set real number value for attribute.
						((XesRealNumberAttribute) attribute).setValue(d);
						break;
					default :
						// Should not happen.
						break;
				}
			}
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			value = attributes.getQName(i);
			if (value.equals(XesXmlTags.KEY) || value.equals(XesXmlTags.VALUE)) {
			} else {
				// Unknown attribute. Raise issue.
				XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(attribute, value);
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

		if (globals != null) {
			for (XesAttribute global : globals) {
				if (global.getKey().equals(attribute.getKey())) {
					// Key already exists in the current container. Raise issue.
					XesXmlIssue issue = new XesXmlIssueAttributeUnique(log, globals, attribute.getKey());
					issue.setLocation(locator);
					issues.add(issue);
				}
			}
			// Add as a global attribute.
			globals.add(attribute);
		} else if (!inValues.isEmpty() && inValues.peek() == 1) {
			// Add as a list value.
			lists.peek().addValue(attribute);
		} else {
			for (XesAttribute otherAttribute : components.peek().getAttributes()) {
				if (otherAttribute.getKey() != null && otherAttribute.getKey().equals(attribute.getKey())) {
					// Key already exists in the current container. Raise issue.
					XesXmlIssue issue = new XesXmlIssueAttributeUnique(components.peek(), attribute.getKey());
					issue.setLocation(locator);
					issues.add(issue);
				}
			}
			// Add as a regular attribute.
			components.peek().addAttribute(attribute);
		}
	}

	/*
	 * Start a values.
	 */
	private void startValuesElement(XesValues values, Attributes attributes) {

		if (lists.isEmpty() || components.isEmpty() || components.peek() != lists.peek()) {
			// Parent must be a list. Raise issue.
			XesXmlIssue issue = new XesXmlIssueParent(values, components.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}
		for (int i = 0; i < attributes.getLength(); i++) {
			// Unknown attribute. Raise issue.
			String value = attributes.getQName(i);
			XesXmlIssue issue = new XesXmlIssueStrictlyConformingAttribute(values, value);
			issue.setLocation(locator);
			issues.add(issue);
		}

		if (!inValues.isEmpty() && inValues.peek() > 0) {
			// Not the first values. Should be only one. Raise issue.
			XesXmlIssue issue = new XesXmlIssueAttributeValues(lists.peek());
			issue.setLocation(locator);
			issues.add(issue);
		}
	}

	/*
	 * Match the classifier keys to the global keys, if possible.
	 */
	private List<String> fixKeys(List<XesAttribute> globals, List<String> keys) {
		List<String> fixedKeys = fixKeys(globals, keys, 0);
		return fixedKeys == null ? keys : fixedKeys;
	}

	/*
	 * Match the remaining classifier keys to the global keys, if possible. Keys
	 * [0,...,index-1] have already been matched.
	 */
	private List<String> fixKeys(List<XesAttribute> globals, List<String> keys, int index) {
		if (index >= keys.size()) {
			// All keys are matched. We're done.
			return keys;
		} else {
			/*
			 * keys[0,...,index-1] are matched to global attributes. Try to
			 * match keys[index].
			 */
			if (findGlobalEventAttribute(globals, keys.get(index))) {
				/*
				 * keys[index] matches a global event attribute. Try if
				 * keys[index+1,..,length-1] match with global event attributes.
				 */
				List<String> fixedKeys = fixKeys(globals, keys, index + 1);
				if (fixedKeys != null) {
					/*
					 * Yes they do. Return the match.
					 */
					return fixedKeys;
				}
				/*
				 * No, they do not. Fall thru to match keys[index]+" "
				 * +keys[index+1] to a global event attribute,
				 */
			}
			/*
			 * No such global event attribute, or no match when key[index] is
			 * matched to a global event attribute. Try merging key[index] with
			 * key[index+1].
			 */
			if (index + 1 == keys.size()) {
				/*
				 * No keys[index+1]. We cannot match keys[length-1]. Fail.
				 */
				return null;
			}
			/*
			 * Copy all matched keys.
			 */
			List<String> newKeys = new ArrayList<String>(keys.size() - 1);
			for (int i = 0; i < index; i++) {
				newKeys.add(keys.get(i));
			}
			/*
			 * Merge keys[index] with keys[index+1].
			 */
			newKeys.add(keys.get(index) + " " + keys.get(index + 1));
			/*
			 * Copy all keys still left to match.
			 */
			for (int i = index + 2; i < keys.size(); i++) {
				newKeys.add(keys.get(i));
			}
			/*
			 * Check match with merged key.
			 */
			return fixKeys(globals, newKeys, index);
		}
	}

	/*
	 * Check whether the given key is in the list of globals.
	 */
	private boolean findGlobalEventAttribute(List<XesAttribute> globals, String key) {
		for (XesAttribute attribute : globals) {
			if (attribute.getKey().equals(key)) {
				// Yes, it is.
				return true;
			}
		}
		// No, it is not.
		return false;
	}

	/*
	 * Returns a date value from the given string value. <null> if it fails to
	 * do so.
	 */
	private Date getDateTime(String value) {
		try {
			Calendar cal = DatatypeConverter.parseDateTime(value);
			return cal.getTime();
		} catch (Exception e) {
			return null;
		}
	}

	/*
	 * Check whether the string value must have a TZD to be parsed successfully
	 * as a date. Returns any valid string if so, and <null> if not so.
	 */
	private String getTZD(String value) {
		int length = value.length();
		if (length < 1) {
			// No TZD possible.
			return null;
		}
		if (value.endsWith("Z")) {
			// "Z" at the end is a valid TZD.
			return "Z";
		}
		if (length < 6) {
			// No TZD possible. Valid values here are strings that end with "-XX:XX" or "+XX:XX",
			// where X denotes any number from 0 to 9.
			return null;
		}
		String plusMinus = value.substring(length - 6, length - 5);
		if (plusMinus.equals("+") || plusMinus.equals("-")) {
			// Found a "+" or "-". OK. 
			return value.substring(length - 6);
		}
		// No "Z" at the end, and no "+" or "-" at the correct position. Cannot be a UTC date.
		return null;
	}

	/*
	 * Returns a decimal number value from the given string value. <null> if it
	 * fails to do so.
	 */
	private Double getDecimal(String value) {
		if (value.contains("e") || value.contains("E")) {
			// Exponents are not allowed.
			return null;
		}
		if (value.contains("NaN") || value.contains("Infinity")) {
			// These values are also not allowed.
			return null;
		}
		try {
			// Any Double should now be a decimal number.
			return Double.valueOf(value);
		} catch (NumberFormatException e) {
			return null;
		}
	}

	/*
	 * Returns a boolean value from the given string value. <null> if it fails
	 * to do so.
	 */
	private Boolean getBoolean(String value) {
		if (value.equals("true") || value.equals("1")) {
			// TRUE
			return true;
		}
		if (value.equals("false") || value.equals("0")) {
			// FALSE
			return false;
		}
		// Not TRUE and not FALSE.
		return null;
	}

	/*
	 * Returns a real number value from the given string value. <null> if it
	 * fails to do so.
	 */
	private Double getDouble(String value) {
		try {
			return Double.valueOf(value);
		} catch (NumberFormatException e) {
			return null;
		}
	}

	/*
	 * Returns a UUID value from the given string value. <null> if it fails to
	 * do so.
	 */
	private UUID getId(String value) {
		try {
			return UUID.fromString(value);
		} catch (IllegalArgumentException e) {
			return null;
		}
	}

	/*
	 * Returns an integer number value from the given string value. <null> if it
	 * fails to do so.
	 */
	private Long getLong(String value) {
		try {
			return Long.valueOf(value);
		} catch (NumberFormatException e) {
			return null;
		}
	}

	/*
	 * Returns an NCName value from the given string value. <null> if it fails
	 * to do so.
	 */
	private String getNCName(String value) {
		return value.contains(":") ? null : value;
	}

	/*
	 * Returns a URI value from the given string value. <null> if it fails to do
	 * so.
	 */
	private URI getURI(String value) {
		try {
			return URI.create(value);
		} catch (IllegalArgumentException e) {
			return null;
		}
	}

	/*
	 * Returns the extension for the attribute with given key. If the prefix of
	 * the key matches the prefix of declared extension, that extension is
	 * returned. If the key has no prefix, noExtension is returned. Otherwise,
	 * <null> is returned.
	 */
	private XesExtension getExtension(String key) {
		String[] values = key.split(":");
		if (values.length <= 1) {
			return noExtension;
		}
		// Check whether the log declares an extension with this prefix.
		XesExtension extension = log.getExtension(values[0]);
		if (extension != null) {
			// Yes it does, return the extension.
			return extension;
		}
		// No it does not. Assume that this is not a prefix of an extension.
		return noExtension;
	}

	/*
	 * Checks whether the given trace has all globals from the given list.
	 */
	private boolean checkGlobals(List<XesAttribute> globals, XesTrace trace) {
		boolean noIssues = true;
		for (XesAttribute global : globals) {
			if (trace.getStringValue(global.getKey()) == null) {
				// Misses this global key. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeGlobal(trace, global);
				issue.setLocation(locator);
				issues.add(issue);
				noIssues = noIssues && !issue.isError(lowestErrorLevel);
			} else if (trace.getType(global.getKey()) != global.getType()) {
				// Has global key, but of wrong type. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeType(trace, global);
				issue.setLocation(locator);
				issues.add(issue);
				noIssues = noIssues && !issue.isError(lowestErrorLevel);
			}
		}
		return noIssues;
	}

	/*
	 * Checks whether the given event has all globals from the given list.
	 */
	private boolean checkGlobals(List<XesAttribute> globals, XesEvent event) {
		boolean noIssues = true;
		for (XesAttribute global : globals) {
			if (event.getStringValue(global.getKey()) == null) {
				// Misses this global key. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeGlobal(event, global);
				issue.setLocation(locator);
				issues.add(issue);
				noIssues = noIssues && !issue.isError(lowestErrorLevel);
			} else if (event.getType(global.getKey()) != global.getType()) {
				// Has global key, but of wrong type. Raise issue.
				XesXmlIssue issue = new XesXmlIssueAttributeType(event, global);
				issue.setLocation(locator);
				issues.add(issue);
				noIssues = noIssues && !issue.isError(lowestErrorLevel);
			}
		}
		return noIssues;
	}

	/*
	 * Returns the log parsed.
	 */
	public XesLog getLog() {
		return log;
	}

	/*
	 * Returns the list of issues raised while parsing the log.
	 */
	public List<XesXmlIssue> getIssues() {
		return issues;
	}
}
